import inspect
import os
from contextlib import nullcontext as does_not_raise

import jax
import numpy as np
import pytest

import nemos as nmo
from nemos.solvers._svrg import SVRG, ProxSVRG, SVRGState
from nemos.third_party.jaxopt import jaxopt
from nemos.tree_utils import pytree_map_and_reduce, tree_l2_norm, tree_slice, tree_sub

# Register every test here as solver-related
pytestmark = pytest.mark.solver_related


@pytest.mark.parametrize(
    ("regr_setup", "stepsize"),
    [
        ("linear_regression", 1e-3),
        ("ridge_regression", 1e-4),
        ("linear_regression_tree", 1e-4),
        ("ridge_regression_tree", 1e-4),
    ],
)
def test_svrg_linear_or_ridge_regression(request, regr_setup, stepsize):
    jax.config.update("jax_enable_x64", True)
    X, y, _, params, loss = request.getfixturevalue(regr_setup)

    param_init = jax.tree_util.tree_map(np.zeros_like, params)
    svrg_params, state = SVRG(loss, tol=10**-12, stepsize=stepsize).run(
        param_init, X, y
    )
    assert pytree_map_and_reduce(
        lambda a, b: np.allclose(a, b, atol=10**-5, rtol=0.0), all, params, svrg_params
    )


@pytest.mark.parametrize(
    "regr_setup",
    [
        "linear_regression",
        "ridge_regression",
        "linear_regression_tree",
        "ridge_regression_tree",
    ],
)
def test_svrg_init_state_default(request, regr_setup):
    jax.config.update("jax_enable_x64", True)
    X, y, _, params, loss = request.getfixturevalue(regr_setup)

    param_init = jax.tree_util.tree_map(np.zeros_like, params)
    svrg = SVRG(loss)
    state = svrg.init_state(param_init, X, y)

    assert state.iter_num == 0
    assert state.key == jax.random.key(123)
    assert state.full_grad_at_reference_point is None
    assert state.reference_point is not None


@pytest.mark.parametrize(
    "regr_setup",
    [
        "linear_regression",
        "ridge_regression",
        "linear_regression_tree",
        "ridge_regression_tree",
    ],
)
def test_svrg_init_state_key(request, regr_setup):
    random_key = jax.random.key(1000)

    jax.config.update("jax_enable_x64", True)
    X, y, _, params, loss = request.getfixturevalue(regr_setup)

    param_init = jax.tree_util.tree_map(np.zeros_like, params)
    svrg = SVRG(loss, key=random_key)
    state = svrg.init_state(param_init, X, y)

    assert state.key == random_key


@pytest.mark.parametrize(
    "regr_setup",
    [
        "linear_regression",
        "linear_regression_tree",
    ],
)
@pytest.mark.parametrize(
    "solver_class, prox, prox_lambda",
    [(SVRG, None, None), (ProxSVRG, jaxopt.prox.prox_ridge, 0.1)],
)
def test_svrg_update_needs_df_xs(request, regr_setup, solver_class, prox, prox_lambda):
    jax.config.update("jax_enable_x64", True)
    X, y, _, params, loss = request.getfixturevalue(regr_setup)

    param_init = jax.tree_util.tree_map(np.zeros_like, params)
    if prox_lambda is not None:
        args = (prox_lambda, X, y)
        constr_args = (loss, prox)
    else:
        args = (X, y)
        constr_args = (loss,)

    solver_class = solver_class(*constr_args)
    state = solver_class.init_state(param_init, *args)

    with pytest.raises(
        ValueError,
        match=r"Full gradient at the anchor point \(state\.full_grad_at_reference_point\) has to be set",
    ):
        _, _ = solver_class.update(param_init, state, *args)


@pytest.mark.parametrize(
    "regularizer_name, solver_class, mask",
    [
        ("Lasso", ProxSVRG, None),
        ("GroupLasso", ProxSVRG, np.array([0, 1, 0, 1]).reshape(1, -1).astype(float)),
        ("Ridge", SVRG, None),
        ("UnRegularized", SVRG, None),
    ],
)
def test_svrg_glm_instantiate_solver(regularizer_name, solver_class, mask):
    solver_name = solver_class.__name__

    # only pass mask if it's not None
    kwargs = {"solver_name": solver_name}
    if mask is not None:
        kwargs["mask"] = mask

    glm = nmo.glm.GLM(
        regularizer=regularizer_name,
        solver_name=solver_name,
        regularizer_strength=None if regularizer_name == "UnRegularized" else 1,
    )
    glm.instantiate_solver()

    # currently glm._solver is a Wrapped(Prox)SVRG
    solver = glm._solver._solver
    assert glm.solver_name == solver_name
    assert isinstance(solver, solver_class)


@pytest.mark.parametrize(
    "regularizer_name, solver_name, mask",
    [
        ("Lasso", "ProxSVRG", None),
        ("GroupLasso", "ProxSVRG", np.array([0, 1, 0, 1]).reshape(1, -1).astype(float)),
        ("Ridge", "SVRG", None),
        ("UnRegularized", "SVRG", None),
    ],
)
@pytest.mark.parametrize("glm_class", [nmo.glm.GLM, nmo.glm.PopulationGLM])
def test_svrg_glm_passes_solver_kwargs(regularizer_name, solver_name, mask, glm_class):
    solver_kwargs = {
        "stepsize": np.abs(np.random.randn()),
        "maxiter": np.random.randint(1, 100),
    }

    # only pass mask if it's not None
    kwargs = {}
    if mask is not None and glm_class == nmo.glm.PopulationGLM:
        kwargs["feature_mask"] = mask

    glm = glm_class(
        regularizer=regularizer_name,
        solver_name=solver_name,
        solver_kwargs=solver_kwargs,
        regularizer_strength=None if regularizer_name == "UnRegularized" else 1,
        **kwargs,
    )
    glm.instantiate_solver()

    # currently glm._solver is a Wrapped(Prox)SVRG
    solver = glm._solver._solver
    assert solver.stepsize == solver_kwargs["stepsize"]
    assert solver.maxiter == solver_kwargs["maxiter"]


@pytest.mark.parametrize(
    "regularizer_name, solver_class, mask",
    [
        ("Lasso", ProxSVRG, None),
        (
            "GroupLasso",
            ProxSVRG,
            np.array([[0.0], [0.0], [1.0]]),
        ),
        ("GroupLasso", ProxSVRG, np.array([[1.0], [0.0], [0.0]])),
        ("Ridge", SVRG, None),
        ("UnRegularized", SVRG, None),
    ],
)
@pytest.mark.parametrize(
    "glm_class",
    [nmo.glm.GLM, nmo.glm.PopulationGLM],
)
def test_svrg_glm_initialize_state(
    glm_class, regularizer_name, solver_class, mask, linear_regression
):
    X, y, _, _, _ = linear_regression

    if glm_class == nmo.glm.PopulationGLM:
        y = np.expand_dims(y, 1)

    reg_cls = getattr(nmo.regularizer, regularizer_name)
    if regularizer_name == "GroupLasso":
        reg = reg_cls(mask=mask)
    else:
        reg = reg_cls()

    # only pass mask if it's not None
    kwargs = {}
    if mask is not None and glm_class == nmo.glm.PopulationGLM:
        kwargs["feature_mask"] = mask

    glm = glm_class(
        regularizer=reg,
        solver_name=solver_class.__name__,
        inverse_link_function=jax.nn.softplus,
        observation_model=nmo.observation_models.PoissonObservations(),
        regularizer_strength=None if regularizer_name == "UnRegularized" else 1,
        **kwargs,
    )

    init_params = glm.initialize_params(X, y)
    state = glm.initialize_state(X, y, init_params)

    assert state.reference_point == init_params

    for f in (glm._solver_init_state, glm._solver_update, glm._solver_run):
        assert isinstance(f.__self__._solver, solver_class)
    assert isinstance(state, SVRGState)


@pytest.mark.parametrize(
    "regularizer_name, solver_class, mask",
    [
        ("Lasso", ProxSVRG, None),
        (
            "GroupLasso",
            ProxSVRG,
            np.array([[0.0], [0.0], [1.0]]),
        ),
        ("Ridge", SVRG, None),
        ("UnRegularized", SVRG, None),
    ],
)
@pytest.mark.parametrize(
    "glm_class",
    [nmo.glm.GLM, nmo.glm.PopulationGLM],
)
def test_svrg_glm_update(
    glm_class, regularizer_name, solver_class, mask, linear_regression
):
    X, y, _, _, loss = linear_regression
    if glm_class == nmo.glm.PopulationGLM:
        y = np.expand_dims(y, 1)

    # only pass mask if it's not None
    kwargs = {}
    if glm_class == nmo.glm.PopulationGLM:
        kwargs["feature_mask"] = mask

    reg_cls = getattr(nmo.regularizer, regularizer_name)
    if regularizer_name == "GroupLasso":
        reg = reg_cls(mask=mask)
    else:
        reg = reg_cls()

    glm = glm_class(
        regularizer=reg,
        solver_name=solver_class.__name__,
        inverse_link_function=jax.nn.softplus,
        observation_model=nmo.observation_models.PoissonObservations(),
        regularizer_strength=None if regularizer_name == "UnRegularized" else 1,
        **kwargs,
    )

    init_params = glm.initialize_params(X, y)
    state = glm.initialize_state(X, y, init_params)

    loss_gradient = jax.jit(jax.grad(glm._solver_loss_fun))

    # initialize full gradient at the anchor point
    state = state._replace(
        full_grad_at_reference_point=loss_gradient(init_params, X, y),
    )

    params, state = glm.update(init_params, state, X, y)

    assert state.iter_num == 1


@pytest.mark.parametrize(
    "regularizer_name, solver_name, mask",
    [
        ("Lasso", "ProxSVRG", None),
        (
            "GroupLasso",
            "ProxSVRG",
            np.array([[0, 1, 0, 1, 1], [1, 0, 1, 0, 0]]).astype(float),
        ),
        ("GroupLasso", "ProxSVRG", np.array([[1, 1, 1, 1, 1]]).astype(float)),
        (
            "GroupLasso",
            "ProximalGradient",
            np.array([[0, 1, 0, 1, 1], [1, 0, 1, 0, 0]]).astype(float),
        ),
        ("GroupLasso", "ProximalGradient", np.array([[1, 1, 1, 1, 1]]).astype(float)),
        ("Ridge", "SVRG", None),
        ("UnRegularized", "SVRG", None),
    ],
)
@pytest.mark.parametrize(
    "maxiter",
    [3, 50],
)
@pytest.mark.parametrize(
    "glm_class",
    [nmo.glm.GLM, nmo.glm.PopulationGLM],
)
def test_svrg_glm_fit(
    glm_class,
    regularizer_name,
    solver_name,
    mask,
    poissonGLM_model_instantiation,
    maxiter,
):
    X, y, model, (w_true, b_true), rate = poissonGLM_model_instantiation

    # set the tolerance such that the solvers never hit their convergence criterion
    # and run until maxiter is reached
    backend = os.getenv("NEMOS_SOLVER_BACKEND")
    if backend is not None:
        use_jaxopt_tol = backend == "jaxopt"
    else:
        use_jaxopt_tol = (
            "jaxopt"
            in str(nmo.solvers._solver_registry.solver_registry[solver_name]).lower()
        )
    tol = -1.0 if use_jaxopt_tol else 0.0
    solver_kwargs = {"maxiter": maxiter, "tol": tol}

    # only pass mask if it's not None
    reg_cls = getattr(nmo.regularizer, regularizer_name)
    if regularizer_name == "GroupLasso":
        reg = reg_cls(mask=mask)
    else:
        reg = reg_cls()

    kwargs = {}
    if glm_class == nmo.glm.PopulationGLM:
        kwargs["feature_mask"] = np.ones((X.shape[1], 1))

    glm = glm_class(
        regularizer=reg,
        solver_name=solver_name,
        inverse_link_function=jax.nn.softplus,
        observation_model=nmo.observation_models.PoissonObservations(),
        solver_kwargs=solver_kwargs,
        regularizer_strength=None if regularizer_name == "UnRegularized" else 1,
        **kwargs,
    )

    if isinstance(glm, nmo.glm.PopulationGLM):
        y = np.expand_dims(y, 1)

    glm.fit(X, y)

    solver = glm._solver
    assert solver.maxiter == maxiter

    assert solver.get_optim_info(glm.solver_state_).num_steps == maxiter


@pytest.mark.parametrize(
    "regularizer_name, solver_class, mask",
    [
        ("Lasso", ProxSVRG, None),
        ("GroupLasso", ProxSVRG, np.array([0, 1, 0]).reshape(-1, 1).astype(float)),
        ("Ridge", SVRG, None),
        ("UnRegularized", SVRG, None),
    ],
)
@pytest.mark.parametrize(
    "glm_class",
    [nmo.glm.GLM, nmo.glm.PopulationGLM],
)
def test_svrg_glm_update_needs_full_grad_at_reference_point(
    glm_class, regularizer_name, solver_class, mask, linear_regression
):
    X, y, _, _, loss = linear_regression
    if glm_class.__name__ == "PopulationGLM":
        y = np.expand_dims(y, 1)

    # only pass mask if it's not None
    reg_cls = getattr(nmo.regularizer, regularizer_name)
    if regularizer_name == "GroupLasso":
        reg = reg_cls(mask=mask)
    else:
        reg = reg_cls()
    kwargs = dict(
        regularizer=reg,
        inverse_link_function=jax.nn.softplus,
        solver_name=solver_class.__name__,
        observation_model=nmo.observation_models.PoissonObservations(),
        regularizer_strength=None if regularizer_name == "UnRegularized" else 0.1,
    )

    if mask is not None and glm_class == nmo.glm.PopulationGLM:
        kwargs["feature_mask"] = np.array([0, 1, 0]).reshape(-1, 1).astype(float)

    glm = glm_class(**kwargs)

    with pytest.raises(
        ValueError,
        match=r"Full gradient at the anchor point \(state\.full_grad_at_reference_point\) has to be set",
    ):
        params = glm.initialize_params(X, y)
        state = glm.initialize_state(X, y, params)
        glm.update(params, state, X, y)


@pytest.mark.parametrize(
    ("regr_setup", "stepsize"),
    [
        ("linear_regression", 1e-3),
        ("ridge_regression", 1e-4),
        ("linear_regression_tree", 1e-4),
        ("ridge_regression_tree", 1e-4),
    ],
)
def test_svrg_update_converges(request, regr_setup, stepsize):
    jax.config.update("jax_enable_x64", True)
    X, y, _, analytical_params, loss = request.getfixturevalue(regr_setup)

    loss_grad = jax.jit(jax.grad(loss))

    N = y.shape[0]
    batch_size = 1
    maxiter = 10_000
    tol = 1e-12
    key = jax.random.key(123)

    m = int((N + batch_size - 1) // batch_size)

    solver = SVRG(loss, stepsize=stepsize, batch_size=batch_size)
    params = jax.tree_util.tree_map(np.zeros_like, analytical_params)
    state = solver.init_state(params, X, y)

    for _ in range(maxiter):
        state = state._replace(
            full_grad_at_reference_point=loss_grad(params, X, y),
        )

        prev_params = params
        for _ in range(m):
            key, subkey = jax.random.split(key)
            ind = jax.random.randint(subkey, (batch_size,), 0, N)
            xi, yi = tree_slice(X, ind), y[ind]
            params, state = solver.update(params, state, xi, yi)

        state = state._replace(
            reference_point=params,
        )

        _error = tree_l2_norm(tree_sub(params, prev_params)) / tree_l2_norm(prev_params)
        if _error < tol:
            break

    assert pytree_map_and_reduce(
        lambda a, b: np.allclose(a, b, atol=10**-5, rtol=0.0),
        all,
        analytical_params,
        params,
    )


@pytest.mark.parametrize(
    "regr_setup, to_tuple",
    [
        ("linear_regression", True),
        ("linear_regression", False),
        ("linear_regression_tree", False),
    ],
)
@pytest.mark.parametrize(
    "prox, prox_lambda",
    [
        (jaxopt.prox.prox_none, None),
        (jaxopt.prox.prox_ridge, 0.1),
        (jaxopt.prox.prox_none, 0.1),
        (nmo.proximal_operator.prox_lasso, 0.1),
    ],
)
def test_svrg_xk_update_step(request, regr_setup, to_tuple, prox, prox_lambda):
    X, y, true_params, ols_coef, loss_arr = request.getfixturevalue(regr_setup)

    # the loss takes an array, but I want to test with tuples as well
    # so make a new loss function that takes a tuple
    if to_tuple:
        true_params = (
            true_params,
            np.zeros(X.shape[1]),
        )
        loss = lambda params, X, y: loss_arr(params[0], X, y)
    else:
        loss = loss_arr

    stepsize = 1e-2
    loss_gradient = jax.jit(jax.grad(loss))

    # set the initial parameters to zero and
    # set the anchor point to a random value that's not just zeros
    init_param = jax.tree_util.tree_map(np.zeros_like, true_params)
    xs = jax.tree_util.tree_map(lambda x: np.random.randn(*x.shape), true_params)
    df_xs = loss_gradient(xs, X, y)

    # sample a mini-batch
    key = jax.random.key(123)
    key, subkey = jax.random.split(key)
    ind = jax.random.randint(subkey, (32,), 0, y.shape[0])
    xi, yi = tree_slice(X, ind), tree_slice(y, ind)

    dfik_xk = loss_gradient(init_param, xi, yi)
    dfik_xs = loss_gradient(xs, xi, yi)

    # update if inputs are arrays
    def _array_update(dfik_xk, dfik_xs, df_xs, init_param, stepsize):
        gk = dfik_xk - dfik_xs + df_xs
        next_xk = init_param - stepsize * gk
        return next_xk

    # update if inputs are a tuple of arrays
    def _tuple_update(dfik_xk, dfik_xs, df_xs, init_param, stepsize):
        return tuple(
            _array_update(a, b, c, d, stepsize)
            for a, b, c, d in zip(dfik_xk, dfik_xs, df_xs, init_param)
        )

    # update if inputs are dicts with either arrays or tuple of arrays as inputs
    # behavior is determined by update_fun
    def _dict_update(dfik_xk, dfik_xs, df_xs, init_param, stepsize, update_fun):
        return {
            k: update_fun(dfik_xk[k], dfik_xs[k], df_xs[k], init_param[k], stepsize)
            for k in dfik_xk.keys()
        }

    if isinstance(true_params, np.ndarray):
        next_xk = _array_update(dfik_xk, dfik_xs, df_xs, init_param, stepsize)
    elif isinstance(true_params, tuple):
        next_xk = _tuple_update(dfik_xk, dfik_xs, df_xs, init_param, stepsize)
    elif isinstance(X, dict):
        assert (
            set(X.keys())
            == set(dfik_xk.keys())
            == set(dfik_xs.keys())
            == set(df_xs.keys())
        )

        if isinstance(list(dfik_xk.values())[0], tuple):
            update_fun = _tuple_update
        else:
            update_fun = _array_update

        next_xk = _dict_update(
            dfik_xk, dfik_xs, df_xs, init_param, stepsize, update_fun
        )
    else:
        raise TypeError

    next_xk = prox(next_xk, prox_lambda, scaling=stepsize)

    if prox_lambda is None:
        assert prox == jaxopt.prox.prox_none
        solver = SVRG(loss)
    else:
        solver = ProxSVRG(loss, prox)
    svrg_next_xk = solver._inner_loop_param_update_step(
        init_param, xs, df_xs, stepsize, prox_lambda, xi, yi
    )

    assert pytree_map_and_reduce(
        lambda a, b: np.allclose(a, b, atol=10**-5, rtol=0.0),
        all,
        next_xk,
        svrg_next_xk,
    )


@pytest.mark.parametrize(
    "shapes, expected_context",
    [
        [
            (10, 10),
            does_not_raise(),
        ],
        [
            (10, 8),
            pytest.raises(
                ValueError,
                match="All arguments must have the same sized first dimension.",
            ),
        ],
    ],
)
def test_svrg_wrong_shapes(shapes, expected_context):
    X = np.random.randn(shapes[0], 3)
    y = np.random.randn(shapes[1], 1)

    init_params = np.random.randn(3, 1)

    def loss_fn(params, X, y):
        return 1.0

    with expected_context:
        svrg = SVRG(loss_fn)
        svrg.run(init_params, X, y)


def test_all_solvers_accept_tol_and_not_atol():
    """All solvers should accept tol and not accept atol."""
    for solver_class in nmo.solvers._solver_registry.solver_registry.values():
        assert "tol" in solver_class.get_accepted_arguments()
        assert "atol" not in solver_class.get_accepted_arguments()


def test_all_solvers_accept_maxiter_and_not_max_steps():
    """All solvers should accept maxiter and not accept max_steps."""
    for solver_class in nmo.solvers._solver_registry.solver_registry.values():
        assert "maxiter" in solver_class.get_accepted_arguments()
        assert "max_steps" not in solver_class.get_accepted_arguments()



# Copyright 2023 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for ZoomLineSearch."""
from contextlib import redirect_stdout
import io

from absl.testing import absltest
from absl.testing import parameterized
import jax
import jax.numpy as jnp
from nemos.third_party.jaxopt.jaxopt import objective
from nemos.third_party.jaxopt.jaxopt import ZoomLineSearch
from nemos.third_party.jaxopt.jaxopt._src import test_util
from nemos.third_party.jaxopt.jaxopt._src.zoom_linesearch import FLAG_CURVATURE_COND_NOT_SATSIFIED
from nemos.third_party.jaxopt.jaxopt._src.zoom_linesearch import FLAG_INTERVAL_NOT_FOUND
from nemos.third_party.jaxopt.jaxopt._src.zoom_linesearch import FLAG_INTERVAL_TOO_SMALL
from nemos.third_party.jaxopt.jaxopt._src.zoom_linesearch import FLAG_NAN_INF_VALUES
from nemos.third_party.jaxopt.jaxopt._src.zoom_linesearch import FLAG_NO_STEPSIZE_FOUND
from nemos.third_party.jaxopt.jaxopt._src.zoom_linesearch import FLAG_NOT_A_DESCENT_DIRECTION
from nemos.third_party.jaxopt.jaxopt.tree_util import tree_add_scalar_mul
from nemos.third_party.jaxopt.jaxopt.tree_util import tree_negative
import numpy as onp
import scipy.optimize
from sklearn import datasets


# pylint: disable=invalid-name
# Uncomment the line below in order to run in float64.
# jax.config.update("jax_enable_x64", True)

class ZoomLinesearchTest(test_util.JaxoptTestCase):
  """Tests for ZoomLineSearch."""

  def setUp(self):
    self.rng = lambda: onp.random.RandomState(0)

  def _check_step_in_state(self, x, p, s, fun, fun_der, state):
    step = tree_add_scalar_mul(x, s, p)
    self.assertAllClose(step, state.params, atol=1e-5, rtol=1e-5)
    self.assertAllClose(fun(step), state.value, atol=1e-5, rtol=1e-5)
    self.assertAllClose(fun_der(step), state.grad, atol=1e-5, rtol=1e-5)
    self.assertTrue(~(state.failed & state.done))

  def _assert_conds(self, s, value_fun, slope_fun, c1=1e-4, c2=0.9, err_msg=""):
    value_init = value_fun(0)
    value_step = value_fun(s)
    slope_init = slope_fun(0)
    slope_step = slope_fun(s)
    msg = (
        "s = {}; value(0) = {}; value(s) = {}; slope(0) = {}; slope(s) = {}; {}"
        .format(s, value_init, value_step, slope_init, slope_step, err_msg)
    )

    self.assertTrue(
        value_step <= value_init + c1 * s * slope_init,
        "Sufficient decrease (Armijo) failed: " + msg,
    )
    self.assertTrue(
        abs(slope_step) <= abs(c2 * slope_init),
        "Small curvature (strong Wolfe) failed: " + msg,
    )

  def _assert_line_conds(self, x, p, s, fun, fun_der, **kw):
    self._assert_conds(
        s,
        value_fun=lambda sp: fun(x + p * sp),
        slope_fun=lambda sp: jnp.dot(fun_der(x + p * sp), p),
        **kw,
    )

  # -- scalar functions

  def _scalar_fun_1(self, s):
    p = -s - s**3 + s**4
    dp = -1 - 3 * s**2 + 4 * s**3
    return p, dp

  def _scalar_fun_2(self, s):
    p = jnp.exp(-4 * s) + s**2
    dp = -4 * jnp.exp(-4 * s) + 2 * s
    return p, dp

  def _scalar_fun_3(self, s):
    p = -jnp.sin(10 * s)
    dp = -10 * jnp.cos(10 * s)
    return p, dp

  # -- n-d functions

  def _line_fun_1(self, x):
    f = jnp.dot(x, x)
    df = 2 * x
    return f, df

  def _line_fun_2(self, x):
    f = jnp.dot(x, jnp.dot(self.A, x)) + 1
    df = jnp.dot(self.A + self.A.T, x)
    return f, df

  def _rosenbrock_fun(self, x):
    return sum(100.0 * (x[1:] - x[:-1] ** 2.0) ** 2.0 + (1 - x[:-1]) ** 2.0)

  def _line_fun_3(self, x):
    # Rosenbrock function
    f = self._rosenbrock_fun(x)
    df = jax.grad(self._rosenbrock_fun)(x)
    return f, df

  # -- Generic scalar searches

  @parameterized.product(
      name=["_scalar_fun_1", "_scalar_fun_2", "_scalar_fun_3"]
  )
  def test_scalar_search(self, name):
    def bind_index(fun, idx):
      # Remember Python's closure semantics!
      return lambda *a, **kw: fun(*a, **kw)[idx]

    value = getattr(self, name)
    fun = bind_index(value, 0)
    fun_der = bind_index(value, 1)
    for old_value in self.rng().randn(3):
      ls = ZoomLineSearch(fun)
      x, p = 0.0, 1.0
      s, state = ls.run(init_stepsize=1.0, params=x, descent_direction=p)
      self._check_step_in_state(x, p, s, fun, fun_der, state)
      self._assert_conds(s, fun, fun_der, err_msg=f"{name} {old_value:g}")

  # -- Generic line searches

  @parameterized.product(name=["_line_fun_1", "_line_fun_2", "_line_fun_3"])
  def test_line_search(self, name):
    def bind_index(fun, idx):
      # Remember Python's closure semantics!
      return lambda *a, **kw: fun(*a, **kw)[idx]

    value = getattr(self, name)
    fun = bind_index(value, 0)
    fun_der = bind_index(value, 1)

    k = 0
    N = 20
    rng = self.rng()
    # sets A in one of the line functions
    self.A = self.rng().randn(N, N)
    while k < 9:
      x = rng.randn(N)
      p = rng.randn(N)
      if jnp.dot(p, fun_der(x)) >= 0:
        # always pick a descent pk
        continue
      if fun(x + 1e6 * p) < fun(x):
        # If the function is unbounded below, the linesearch cannot finish
        continue
      k += 1

      f0 = fun(x)
      g0 = fun_der(x)
      ls = ZoomLineSearch(fun)
      s, state = ls.run(
          init_stepsize=1.0, params=x, descent_direction=p, value=f0, grad=g0
      )
      self._check_step_in_state(x, p, s, fun, fun_der, state)
      self._assert_line_conds(x, p, s, fun, fun_der, err_msg=f"{name}")

  def test_logreg(self):
    x, y = datasets.make_classification(
        n_samples=10, n_features=5, n_classes=2, n_informative=3, random_state=0
    )
    data = (x, y)
    fun = objective.binary_logreg

    def fun_(w):
      return fun(w, data)

    rng = onp.random.RandomState(0)
    w_init = rng.randn(x.shape[1])
    initial_grad = jax.grad(fun)(w_init, data=data)
    descent_dir = tree_negative(initial_grad)

    # Call to run.
    ls = ZoomLineSearch(fun=fun, maxiter=20)
    stepsize, state = ls.run(
        init_stepsize=1.0, params=w_init, fun_kwargs={"data": data}
    )

    self._assert_line_conds(
        w_init, descent_dir, stepsize, fun_, jax.grad(fun_), c1=ls.c1, c2=ls.c2
    )
    self._check_step_in_state(
        w_init, descent_dir, stepsize, fun_, jax.grad(fun_), state
    )

    # Call to run with value_and_grad=True
    ls = ZoomLineSearch(
        fun=jax.value_and_grad(fun), maxiter=20, value_and_grad=True
    )
    stepsize, state = ls.run(
        init_stepsize=1.0, params=w_init, fun_kwargs={"data": data}
    )

    self._assert_line_conds(
        w_init, descent_dir, stepsize, fun_, jax.grad(fun_), c1=ls.c1, c2=ls.c2
    )
    self._check_step_in_state(
        w_init, descent_dir, stepsize, fun_, jax.grad(fun_), state
    )

  def test_failure_descent_direction(self):
    # See gh-7475

    # For this f and p, starting at a point on axis 0, the strong Wolfe
    # condition 2 is met if and only if the step length s satisfies
    # |x + s| <= c2 * |x|
    def fun(x):
      return jnp.dot(x, x)

    p = jnp.array([1.0, 0.0])
    x = 60 * p

    # Test that the line search fails for p not a descent direction
    # For high maxiter, still finds a decrease error because of 
    # the approximate Wolfe condition so we reduced maxiter
    ls = ZoomLineSearch(fun, c2=0.5, maxiter=18, verbose=True)
    stdout = io.StringIO()
    with redirect_stdout(stdout):
      s, state = ls.run(init_stepsize=1.0, params=x, descent_direction=p)
    self._check_step_in_state(x, p, s, fun, jax.grad(fun), state)
    # Check that we were not able to make a step or an infinitesimal one
    self.assertLess(s, 1e-5)
    self.assertTrue(FLAG_NOT_A_DESCENT_DIRECTION in stdout.getvalue())
    self.assertTrue(FLAG_NO_STEPSIZE_FOUND in stdout.getvalue())

  def test_failure_too_small_max_stepsize(self):
    def fun(x):
      return jnp.dot(x, x)

    p = jnp.array([1.0, 0.0])
    x = -60 * p

    # Test that the line search fails if the maximum stepsize is too small
    # Here, smallest s satisfying strong Wolfe conditions for c2=0.5 is 30
    ls = ZoomLineSearch(fun, c2=0.5, max_stepsize=10, verbose=True)
    stdout = io.StringIO()
    with redirect_stdout(stdout):
      s, state = ls.run(init_stepsize=1.0, params=x, descent_direction=p)
    self._check_step_in_state(x, p, s, fun, jax.grad(fun), state)
    # Check that we still made a step
    self.assertTrue(s == 10.0)
    self.assertTrue(FLAG_INTERVAL_NOT_FOUND in stdout.getvalue())
    self.assertTrue(FLAG_CURVATURE_COND_NOT_SATSIFIED in stdout.getvalue())

  def test_failure_not_enough_iter(self):
    def fun(x):
      return jnp.dot(x, x)

    p = jnp.array([1.0, 0.0])
    x = -60 * p

    c2 = 0.5
    # s=30 will only be tried on the 6th iteration, so this fails because
    # the maximum number of iterations is reached.
    ls = ZoomLineSearch(fun, c2=c2, maxiter=5, verbose=True)
    stdout = io.StringIO()
    with redirect_stdout(stdout):
      s, state = ls.run(init_stepsize=1.0, params=x, descent_direction=p)
    self._check_step_in_state(x, p, s, fun, jax.grad(fun), state)
    # Check that we still made a step
    self.assertTrue(s == 16.0)
    self.assertTrue(state.failed)
    # Here the error should not be that we haven't had a descent direction
    self.assertFalse(FLAG_NOT_A_DESCENT_DIRECTION in stdout.getvalue())

    # Check if it works normally
    ls = ZoomLineSearch(fun, c2=c2)
    s, state = ls.run(init_stepsize=1.0, params=x, descent_direction=p)
    self._assert_line_conds(x, p, s, fun, jax.grad(fun), c1=ls.c1, c2=c2)
    self._check_step_in_state(x, p, s, fun, jax.grad(fun), state)
    self.assertTrue(s >= 30.0)

  def test_failure_flat_fun(self):
    # Check failure for a very flat function
    def fun_flat(x):
      return jnp.exp(-1 / x**2)

    x = jnp.asarray(-0.2)
    if x.dtype == "float64":
      x = x / 2.0
    ls = ZoomLineSearch(fun_flat, verbose=True)
    stdout = io.StringIO()
    with redirect_stdout(stdout):
      ls.run(init_stepsize=1.0, params=x)
    self.assertTrue(FLAG_INTERVAL_TOO_SMALL in stdout.getvalue())

  def test_failure_inf_value(self):
    # Check behavior for inf/nan values
    def fun_inf(x):
      return jnp.log(x)

    x = 1.0
    p = -2.0
    ls = ZoomLineSearch(fun_inf, verbose=True)
    stdout = io.StringIO()
    with redirect_stdout(stdout):
      ls.run(init_stepsize=1.0, params=x, descent_direction=p)
    self.assertTrue(FLAG_NAN_INF_VALUES in stdout.getvalue())

  def test_aux_value(self):
    def fun(x):
      return jnp.cos(jnp.sum(jnp.exp(-x)) ** 2), x

    x = jnp.ones(2)
    p = jnp.array([-0.5, -0.25])
    ls = ZoomLineSearch(fun=fun, maxiter=100, has_aux=True)
    new_stepsize, state = ls.run(
        init_stepsize=1.0, params=x, descent_direction=p
    )
    new_x = x + new_stepsize * p
    self.assertArraysEqual(state.aux, new_x)

  def test_against_scipy(self):
    def fun(x):
      return jnp.cos(jnp.sum(jnp.exp(-x)) ** 2)

    x = jnp.ones(2)
    p = jnp.array([-0.5, -0.25])

    ls = ZoomLineSearch(fun=fun, maxiter=20)
    stepsize, state = ls.run(init_stepsize=1.0, params=x, descent_direction=p)

    ls2 = ZoomLineSearch(
        fun=jax.value_and_grad(fun),
        maxiter=20,
        value_and_grad=True,
    )
    stepsize2, state2 = ls2.run(
        init_stepsize=1.0, params=x, descent_direction=p
    )

    scipy_res = scipy.optimize.line_search(fun, jax.grad(fun), x, p)

    self.assertAllClose(scipy_res[0], stepsize, atol=1e-5, check_dtypes=False)
    self.assertAllClose(
        scipy_res[3], state.value, atol=1e-5, check_dtypes=False
    )

    self.assertAllClose(scipy_res[0], stepsize2, atol=1e-5, check_dtypes=False)
    self.assertAllClose(
        scipy_res[3], state2.value, atol=1e-5, check_dtypes=False
    )

  def test_high_smaller_than_low(self):
    # See google/jax/issues/16236
    def fun(x):
      return x**2

    # Descent direction p chosen such that, with x+p
    # the first trial of the algorithm,
    # 1. p*f'(x) < 0 (valid descent direction)
    # 2. x+p satisifies sufficient decrease
    # 3. x+p does not satisfy small curvature
    # 4. f'(x+p) > 0
    # As a result, the first trial starts with high < low

    x = -1.0
    p = -1.95 * x

    ls = ZoomLineSearch(fun)
    _, state = ls.run(init_stepsize=1.0, params=x, descent_direction=p)

    self.assertFalse(state.failed)

  @parameterized.product(out_dtype=[jnp.float32, jnp.float64])
  def test_correct_dtypes(self, out_dtype):
    def fun(x):
      return jnp.cos(jnp.sum(jnp.exp(-x)) ** 2).astype(out_dtype)

    with jax.experimental.enable_x64():
      xk = jnp.ones(2, dtype=jnp.float32)
      pk = jnp.array([-0.5, -0.25], dtype=jnp.float32)
      ls = ZoomLineSearch(fun, maxiter=100)
      _, state = ls.run(init_stepsize=1.0, params=xk, descent_direction=pk)
      for name in ("done", "interval_found", "failed"):
        self.assertEqual(getattr(state, name).dtype, jnp.bool_)
      for name in ("iter_num", ):
        self.assertEqual(getattr(state, name).dtype, jnp.int64)
      for name in ("num_fun_eval", "num_grad_eval"):
        self.assertEqual(getattr(state, name).dtype, jnp.int32)
      for name in (
          "params",
          "grad",
          "descent_direction",
          "slope_init",
          "prev_slope_step",
          "slope_low",
          "slope_high",
      ):
        self.assertEqual(getattr(state, name).dtype, jnp.float32, name)
      for name in ("prev_stepsize", "low", "high", "cubic_ref"):
        self.assertEqual(getattr(state, name).dtype, jnp.float32, name)
      for name in (
          "value",
          "value_init",
          "prev_value_step",
          "value_low",
          "value_high",
          "value_cubic_ref",
          "error",
      ):
        self.assertEqual(getattr(state, name).dtype, out_dtype)

  @parameterized.product(jit=[False, True])
  def test_non_jittable(self, jit):
    def fun(x):
      return -onp.sin(10 * x), -10 * onp.cos(10 * x)

    x = 1.0

    def run_ls():
      ls = ZoomLineSearch(fun, value_and_grad=True, jit=jit)
      ls.run(init_stepsize=1.0, params=x)

    if jit:
      self.assertRaises(jax.errors.TracerArrayConversionError, run_ls)
    else:
      run_ls()


if __name__ == "__main__":
  absltest.main()



"""
Inspect utility function module.

This module provides utilities for inspecting class hierarchies,
abstract methods, and subclass method implementations.

Modules
-------
inspect_utils : module
    Contains utility functions to analyze abstract and concrete class methods,
    identify abstract classes, and verify method compliance in subclasses.
"""

from .inspect_utils import (
    check_all_abstract_methods_compliance,
    get_abstract_classes,
    get_non_abstract_classes,
    get_subclass_methods,
    get_superclass_abstract_methods,
    is_abstract,
    list_abstract_methods,
    reimplements_method,
    trim_kwargs,
)

__all__ = [
    "reimplements_method",
    "get_subclass_methods",
    "list_abstract_methods",
    "is_abstract",
    "get_non_abstract_classes",
    "get_abstract_classes",
    "get_superclass_abstract_methods",
    "check_all_abstract_methods_compliance",
    "trim_kwargs",
]



# Copyright 2021 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""SGD solver with Polyak step size."""

import dataclasses
from typing import Any
from typing import Callable
from typing import NamedTuple
from typing import Optional
from typing import Union

import jax.numpy as jnp

from nemos.third_party.jaxopt.jaxopt._src import base
from nemos.third_party.jaxopt.jaxopt.tree_util import tree_add
from nemos.third_party.jaxopt.jaxopt.tree_util import tree_add_scalar_mul
from nemos.third_party.jaxopt.jaxopt.tree_util import tree_l2_norm
from nemos.third_party.jaxopt.jaxopt.tree_util import tree_scalar_mul
from nemos.third_party.jaxopt.jaxopt.tree_util import tree_sub
from nemos.third_party.jaxopt.jaxopt.tree_util import tree_zeros_like
from nemos.third_party.jaxopt.jaxopt._src.tree_util import tree_single_dtype


class PolyakSGDState(NamedTuple):
  """Named tuple containing state information."""
  iter_num: int
  error: float
  value: float
  aux: Any
  stepsize: float
  velocity: Optional[Any]

  num_fun_eval: int = 0
  num_grad_eval: int = 0


@dataclasses.dataclass(eq=False)
class PolyakSGD(base.StochasticSolver):
  r"""SGD with Polyak step size.

  The stochastic Polyak step-size is a simple and efficient step-size for SGD.
  Although this algorithm does not require to set a step-size parameter, it does
  require knowledge of a lower bound on the objective function (see below).
  Furthermore, some variants accept other hyperparameters. 

  .. warning::
      This method requires knowledge of an approximate value of the of the objective function
      minimum, passed through the ``fun_min`` argument. For overparametrized models, this can be
      set to 0 (default value). Failing to set an appropriate value for ``fun_min`` can lead to
      a model that diverges or converges to a suboptimal solution.


  This class implements two different variants of the stochastic Polyak step size method: ``SPS_max``
  and ``SPS+``. The ``SPS_max`` variant from (Loizou et al. 2021) accepts the hyperparameters
  ``max_stepsize`` and ``delta`` and sets the current step-size :math:`\gamma`  as

  .. math::

    \gamma = \min\left\{\frac{\text{fun}(x) - \text{fun}(x^\star)}{\|\nabla \text{fun}(x)\|^2 + \text{delta}}, \text{max_stepsize}\right\}

  while for the ``SPS+`` variant, it is given by

  .. math::

    \gamma = \max\left\{0, \frac{\text{fun}(x) - \text{fun}(x^\star)}{\|\nabla \text{fun}(x)\|^2}\right\}

  and the step-size is zero whenever :math:`\|\nabla \text{fun}(x)\|^2` is zero.

  In all cases, the step size is then used in the update

  .. math::

    v_{t+1} &= \text{momentum} v_t - \gamma \nabla \text{fun}(x) \\
    x_{t+1} &= x_t + v_{t+1}


  Attributes:
    fun: a function of the form ``fun(params, *args, **kwargs)``, where
      ``params`` are parameters of the model,
      ``*args`` and ``**kwargs`` are additional arguments.
    value_and_grad: whether ``fun`` just returns the value (False) or both the
      value and gradient (True).
    has_aux: whether ``fun`` outputs auxiliary data or not.
      If ``has_aux`` is False, ``fun`` is expected to be
        scalar-valued.
      If ``has_aux`` is True, then we have one of the following
        two cases.
      If ``value_and_grad`` is False, the output should be
      ``value, aux = fun(...)``.
      If ``value_and_grad == True``, the output should be
      ``(value, aux), grad = fun(...)``.
      At each iteration of the algorithm, the auxiliary outputs are stored
        in ``state.aux``.
    fun_min: a lower bound on fun.

    variant: which version of the stochastic Polyak step-size is implemented.
      Can be one of "SPS_max" or "SPS+".
    max_stepsize: a maximum step size to use. Only used when variant="SPS_max".
    delta: a value to add in the denominator of the update (default: 0).
    momentum: momentum parameter, 0 corresponding to no momentum.
    pre_update: a function to execute before the solver's update.
      The function signature must be
      ``params, state = pre_update(params, state, *args, **kwargs)``.

    maxiter: maximum number of solver iterations.
    tol: tolerance to use.
    verbose: whether to print information on every iteration or not.

    implicit_diff: whether to enable implicit diff or autodiff of unrolled
      iterations.
    implicit_diff_solve: the linear system solver to use.

    jit: whether to JIT-compile the optimization loop (default: True).
    unroll: whether to unroll the optimization loop (default: "auto").

  References:
    Berrada, Leonard and Zisserman, Andrew and Kumar, M Pawan.
    `"Training neural networks for and by interpolation" <https://arxiv.org/abs/1906.05661>`_.
    International Conference on Machine Learning, 2020.


    Loizou, Nicolas and Vaswani, Sharan and Laradji, Issam Hadj and
    Lacoste-Julien, Simon.
    `"Stochastic polyak step-size for sgd: An adaptive learning rate for fast
    convergence" <https://arxiv.org/abs/2002.10542>`_.
    International Conference on Artificial Intelligence and Statistics, 2021.
  """
  fun: Callable
  value_and_grad: bool = False
  has_aux: bool = False
  fun_min: float = 0.0

  variant: str = "SPS_max"
  max_stepsize: float = 1.0
  delta: float = 0.0
  momentum: float = 0.0
  pre_update: Optional[Callable] = None

  maxiter: int = 500
  tol: float = 1e-3
  verbose: Union[bool, int] = False

  implicit_diff: bool = False
  implicit_diff_solve: Optional[Callable] = None

  jit: bool = True
  unroll: base.AutoOrBoolean = "auto"

  def init_state(self,
                 init_params: Any,
                 *args,
                 **kwargs) -> PolyakSGDState:
    """Initialize the solver state.

    Args:
      init_params: pytree containing the initial parameters.
      *args: additional positional arguments to be passed to ``fun``.
      **kwargs: additional keyword arguments to be passed to ``fun``.
    Returns:
      state
    """
    value, aux = self._fun(init_params, *args, **kwargs)

    if self.momentum == 0:
      velocity = None
    else:
      velocity = tree_zeros_like(init_params)

    params_dtype = tree_single_dtype(init_params)

    return PolyakSGDState(iter_num=jnp.asarray(0),
                          error=jnp.asarray(jnp.inf, dtype=params_dtype),
                          value=jnp.asarray(jnp.inf, dtype=value.dtype),
                          stepsize=jnp.asarray(1.0, dtype=params_dtype),
                          aux=aux,
                          velocity=velocity,
                          num_fun_eval=jnp.array(1, base.NUM_EVAL_DTYPE),
                          num_grad_eval=jnp.array(0, base.NUM_EVAL_DTYPE))

  def update(self,
             params: Any,
             state: PolyakSGDState,
             *args,
             **kwargs) -> base.OptStep:
    """Performs one iteration of the solver.

    Args:
      params: pytree containing the parameters.
      state: named tuple containing the solver state.
      *args: additional positional arguments to be passed to ``fun``.
      **kwargs: additional keyword arguments to be passed to ``fun``.
    Returns:
      (params, state)
    """
    dtype = tree_single_dtype(params)

    if self.pre_update:
      params, state = self.pre_update(params, state, *args, **kwargs)

    (value, aux), grad = self._value_and_grad_fun(params, *args, **kwargs)

    grad_sqnorm = tree_l2_norm(grad, squared=True)
    if self.variant == "SPS_max":
      stepsize = jnp.minimum(
          (value - self.fun_min) / (grad_sqnorm + self.delta), self.max_stepsize
      )
    elif self.variant == "SPS+":
      # if grad_sqnorm is smaller than machine epsilon, we set the stepsize to 0
      stepsize = jnp.where(
          grad_sqnorm <= jnp.finfo(dtype).eps,
          0.0,
          jnp.maximum((value - self.fun_min) / grad_sqnorm, 0),
      )
    else:
      raise NotImplementedError(f"Unknown variant {self.variant}")

    stepsize = stepsize.astype(state.stepsize.dtype)

    if self.momentum == 0:
      new_params = tree_add_scalar_mul(params, -stepsize, grad)
      new_velocity = None
    else:
      # new_v = momentum * v - step_size * grad
      # new_params = params + new_v
      new_velocity = tree_sub(tree_scalar_mul(self.momentum, state.velocity),
                              tree_scalar_mul(stepsize, grad))
      new_params = tree_add(params, new_velocity)

    error = jnp.sqrt(grad_sqnorm)
    new_state = PolyakSGDState(iter_num=state.iter_num + 1,
                               error=jnp.asarray(error, dtype=dtype),
                               velocity=new_velocity,
                               value=jnp.asarray(value),
                               stepsize=jnp.asarray(stepsize, dtype=dtype),
                               aux=aux,
                               num_fun_eval=state.num_fun_eval + 1,
                               num_grad_eval=state.num_grad_eval + 1)

    if self.verbose:
      self.log_info(
          new_state,
          error_name="Gradient Norm",
          additional_info={
              "Objective Value": value,
              "Stepsize": stepsize,
          }
      )
    return base.OptStep(params=new_params, state=new_state)

  def optimality_fun(self, params, *args, **kwargs):
    """Optimality function mapping compatible with ``@custom_root``."""
    return self._grad_fun(params, *args, **kwargs)[0]

  def __hash__(self):
    # We assume that the attribute values completely determine the solver.
    return hash(self.attribute_values())

  def __post_init__(self):
    super().__post_init__()

    self._fun, self._grad_fun, self._value_and_grad_fun = (
        base._make_funs_with_aux(
            fun=self.fun,
            value_and_grad=self.value_and_grad,
            has_aux=self.has_aux,
        )
    )

    self.reference_signature = self.fun
